import { EntryIndex } from "./entry-index.js";
import * as Sorting from "./log-sorting.js";
import { EntryFetchAllOptions, EntryFetchOptions } from "./entry-io.js";
import { EncryptionTemplateMaybeEncrypted, Entry, Payload, CanAppend } from "./entry.js";
import { HLC, LamportClock, Timestamp } from "./clock.js";
import { PublicKeyEncryptionResolver, SignatureWithKey } from "@dao-xyz/peerbit-crypto";
import { Encoding } from "./encoding.js";
import { Identity } from "./identity.js";
import { HeadsIndex } from "./heads.js";
import { BlockStore } from "@dao-xyz/libp2p-direct-block";
import { Values } from "./values.js";
export type TrimToLengthOption = {
    type: "length";
    to: number;
    from?: number;
};
export type TrimToByteLengthOption = {
    type: "bytelength";
    to: number;
    from?: number;
};
export type TrimCondition = TrimToByteLengthOption | TrimToLengthOption;
export type TrimOptions = {
    canTrim?: (entry: Entry<any>) => Promise<boolean> | boolean;
} & TrimCondition;
export type Change<T> = {
    added: Entry<T>[];
    removed: Entry<T>[];
};
export type LogOptions<T> = {
    encryption?: PublicKeyEncryptionResolver;
    encoding?: Encoding<T>;
    logId?: string;
    entries?: Entry<T>[];
    heads?: any;
    clock?: LamportClock;
    sortFn?: Sorting.ISortFunction;
    concurrency?: number;
    trim?: TrimOptions;
};
/**
 * @description
 * Log implements a G-Set CRDT and adds ordering.
 *
 * From:
 * "A comprehensive study of Convergent and Commutative Replicated Data Types"
 * https://hal.inria.fr/inria-00555588
 */
export declare class Log<T> {
    _sortFn: Sorting.ISortFunction;
    _storage: BlockStore;
    _id: string;
    _hlc: HLC;
    _identity: Identity;
    _entryIndex: EntryIndex<T>;
    _headsIndex: HeadsIndex<T>;
    _values: Values<T>;
    _nextsIndex: Map<string, Set<string>>;
    _trim?: TrimOptions;
    _encryption?: PublicKeyEncryptionResolver;
    _encoding: Encoding<T>;
    joinConcurrency: number;
    constructor(store: BlockStore, identity: Identity, options?: LogOptions<T>);
    /**
     * Returns the length of the log.
     */
    get length(): number;
    /**
     * Get all entries sorted
     */
    get values(): Entry<T>[];
    /**
     * Returns the head index
     */
    get headsIndex(): HeadsIndex<T>;
    /**
     * Don't use this anywhere performance matters
     */
    get heads(): Entry<T>[];
    /**
     * Returns an array of Entry objects that reference entries which
     * are not in the log currently.
     * @returns {Array<Entry<T>>}
     */
    get tails(): Entry<T>[];
    /**
     * Returns an array of hashes that are referenced by entries which
     * are not in the log currently.
     * @returns {Array<string>} Array of hashes
     */
    get tailHashes(): string[];
    /**
     * Get local HLC
     */
    get hlc(): HLC;
    /**
     * Set the identity for the log
     * @param {Identity} [identity] The identity to be set
     */
    setIdentity(identity: Identity): void;
    /**
     * Find an entry.
     * @param {string} [hash] The hashes of the entry
     * @returns {Entry|undefined}
     */
    get(hash: string): Entry<T> | undefined;
    /**
     * Checks if a entry is part of the log
     * @param {string} hash The hash of the entry
     * @returns {boolean}
     */
    has(entry: Entry<string> | string): boolean;
    traverse(rootEntries: Entry<T>[], amount?: number, endHash?: string): {
        [key: string]: Entry<T>;
    };
    getReferenceSamples(from: Entry<T>, options?: {
        pointerCount?: number;
        memoryLimit?: number;
    }): Entry<T>[];
    /**
     * Append an entry to the log.
     * @param {Entry} entry Entry to add
     * @return {Log} New Log containing the appended value
     */
    append(data: T, options?: {
        canAppend?: CanAppend<T>;
        gidSeed?: Uint8Array;
        nexts?: Entry<any>[];
        identity?: Identity;
        signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
        reciever?: EncryptionTemplateMaybeEncrypted;
        onGidsShadowed?: (gids: string[]) => void;
        trim?: TrimOptions;
        timestamp?: Timestamp;
    }): Promise<{
        entry: Entry<T>;
        removed: Entry<T>[];
    }>;
    iterator(options?: {
        from?: "tail" | "head";
        amount?: number;
    }): IterableIterator<Entry<T>>;
    /**
     * Join two logs.
     *
     * Joins another log into this one.
     *
     * @param {Log} log Log to join with this Log
     * @param {number} [size=-1] Max size of the joined log
     * @returns {Promise<Log>} This Log instance
     * @example
     * await log1.join(log2)
     */
    join(log: Log<T>, options?: {
        verifySignatures?: boolean;
        trim?: TrimOptions;
    }): Promise<Change<T>>;
    /**
     * @param options
     * @returns deleted entries
     */
    trim(option?: TrimOptions | undefined): Promise<Entry<T>[]>;
    deleteRecursively(from: Entry<any> | Entry<any>[]): Promise<void>;
    delete(entry: Entry<any>): Promise<void>;
    /**
     * Get the log in JSON format.
     * @returns {Object} An object with the id and heads properties
     */
    toJSON(): {
        id: string;
        heads: string[];
    };
    /**
     * Get the log in JSON format as a snapshot.
     * @returns {Object} An object with the id, heads and value properties
     */
    toSnapshot(): {
        id: string;
        heads: Entry<T>[];
        values: Entry<T>[];
    };
    /**
     * Returns the log entries as a formatted string.
     * @returns {string}
     * @example
     * two
     * └─one
     *   └─three
     */
    toString(payloadMapper?: (payload: Payload<T>) => string): string;
    /**
     * Get the log's multihash.
     * @returns {Promise<string>} Multihash of the Log as Base58 encoded string.
     */
    toMultihash(options?: {
        format?: string;
    }): Promise<string>;
    static fromMultihash<T>(store: BlockStore, identity: Identity, hash: string, options: {
        sortFn?: Sorting.ISortFunction;
    } & EntryFetchAllOptions<T>): Promise<Log<T>>;
    static fromEntryHash<T>(store: BlockStore, identity: Identity, hash: string | string[], options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        logId?: string;
        length?: number;
        exclude?: any[];
        shouldFetch?: (hash: string) => boolean;
        timeout?: number;
        concurrency?: number;
        sortFn?: any;
        onFetched?: any;
    }): Promise<Log<T>>;
    /**
     * Create a log from a Log Snapshot JSON.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {Object} json Log snapshot as JSON object
     * @param {Object} options
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many entries to include in the log
     * @param {function(hash, entry,  parent, depth)} [options.onFetched]
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @return {Promise<Log>} New Log
     */
    static fromJSON<T>(store: BlockStore, identity: Identity, json: {
        id: string;
        heads: string[];
    }, options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        length?: number;
        timeout?: number;
        sortFn?: Sorting.ISortFunction;
        onFetched?: (entry: Entry<T>) => void;
    }): Promise<Log<T>>;
    static fromEntry<T>(store: BlockStore, identity: Identity, sourceEntries: Entry<T>[] | Entry<T>, options: EntryFetchOptions<T> & {
        shouldFetch?: (hash: string) => boolean;
        encryption?: PublicKeyEncryptionResolver;
        sortFn?: Sorting.ISortFunction;
    }): Promise<Log<T>>;
    /**
     * Find heads from a collection of entries.
     *
     * Finds entries that are the heads of this collection,
     * ie. entries that are not referenced by other entries.
     *
     * @param {Array<Entry<T>>} entries Entries to search heads from
     * @returns {Array<Entry<T>>}
     */
    static findHeads<T>(entriesOrLogs: (Entry<T> | Log<T>)[]): Entry<T>[];
    static findTails<T>(entries: Entry<T>[]): Entry<T>[];
    static findTailHashes(entries: Entry<any>[]): string[];
    static difference<T>(from: Log<T>, into: Log<T>): Promise<Map<string, Entry<T>>>;
}
